"""Datasets and related utilities."""

import dataclasses

from typing import List, Optional

import torch
from torch import nn
from torch.utils import data

from . import mappers, tsv


class Item(nn.Module):
    """Source tensor, with optional features and target tensors.

    This represents a single item or observation.

    Args:
        source (torch.Tensor).
        features (torch.Tensor, optional).
        target (torch.Tensor, optional).
    """

    source: torch.Tensor
    features: Optional[torch.Tensor]
    target: Optional[torch.Tensor]

    def __init__(self, source, features=None, target=None):
        super().__init__()
        self.register_buffer("source", source)
        self.register_buffer("features", features)
        self.register_buffer("target", target)

    @property
    def has_features(self) -> bool:
        return self.features is not None

    @property
    def has_target(self) -> bool:
        return self.target is not None


# TODO: Add an iterable data set object for out-of-core inference.


@dataclasses.dataclass
class Dataset(data.Dataset):
    """Mappable data set.

    This class loads the entire file into memory and is therefore only suitable
    for in-core data sets.
    """

    samples: List[tsv.SampleType]
    mapper: mappers.Mapper
    parser: tsv.TsvParser  # Ditto.

    # Properties.

    @property
    def has_features(self) -> bool:
        return self.parser.has_features

    @property
    def has_target(self) -> bool:
        return self.parser.has_target

    # Required API.

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, idx: int) -> Item:
        """Retrieves item by index.

        Args:
            idx (int).

        Returns:
            Item.
        """
        sample = self.samples[idx]
        if self.has_features:
            if self.has_target:
                source, features, target = sample
                return Item(
                    source=self.mapper.encode_source(source),
                    features=self.mapper.encode_features(features),
                    target=self.mapper.encode_target(target),
                )
            else:
                source, features = sample
                return Item(
                    source=self.mapper.encode_source(source),
                    features=self.mapper.encode_features(features),
                )
        elif self.has_target:
            source, target = sample
            return Item(
                source=self.mapper.encode_source(source),
                target=self.mapper.encode_target(target),
            )
        else:
            source = sample
            return Item(source=self.mapper.encode_source(source))



"""Batching, padding, and related utilities.

Anything which has a tensor member should inherit from nn.Module, run the
superclass constructor, and register the tensor as a buffer. This enables the
Trainer to move them to the appropriate device."""

from typing import Callable, List, Optional

import torch
from torch import nn

from .. import special


class PaddedTensor(nn.Module):
    """A tensor with an optional padding mask.

    This is ordinarily used for padding a tensor list, so it represents
    one of (source, target, features) for a batch.

    Mask and string length tensors can be generated as needed.

    Args:
        tensorlist (List[torch.Tensor]): a list of tensors.
        length_msg_callback (Callable[[int], None]): callback for handling a
            violation of expected tensor length.
    """

    pad_idx: int
    padded: torch.Tensor

    def __init__(
        self,
        tensorlist: List[torch.Tensor],
        length_msg_callback: Optional[Callable[[int], None]] = None,
    ):
        super().__init__()
        pad_len = max(len(tensor) for tensor in tensorlist)
        if length_msg_callback is not None:
            length_msg_callback(pad_len)
        self.register_buffer(
            "padded",
            torch.stack(
                [self.pad_tensor(tensor, pad_len) for tensor in tensorlist],
            ),
        )

    @property
    def mask(self) -> torch.Tensor:
        return self.padded == special.PAD_IDX

    @staticmethod
    def pad_tensor(tensor: torch.Tensor, pad_max: int) -> torch.Tensor:
        """Pads a tensor.

        Args:
            tensor (torch.Tensor).
            pad_max (int): desired tensor length.

        Returns:
            torch.Tensor.
        """
        padding = pad_max - len(tensor)
        return nn.functional.pad(
            tensor, (0, padding), "constant", special.PAD_IDX
        )

    def __len__(self) -> int:
        return len(self.padded)

    def lengths(self) -> torch.Tensor:
        """Computes the lengths of all the strings in the tensor.

        This needs to be on CPU for packing.

        Returns:
            torch.Tensor.
        """
        return (~self.mask).sum(dim=1).cpu()


class Batch(nn.Module):
    """Padded source tensor, with optional padded features and target tensors.

    This represents a padded batch. It is produced by the collator and fed to
    the trainer.

    Args:
        source (torch.Tensor).
        features (torch.Tensor, optional).
        target (torch.Tensor, optional).
    """

    source: PaddedTensor
    features: Optional[PaddedTensor]
    target: Optional[PaddedTensor]

    def __init__(self, source, features=None, target=None):
        super().__init__()
        self.register_module("source", source)
        self.register_module("features", features)
        self.register_module("target", target)

    @property
    def has_features(self) -> bool:
        return self.features is not None

    @property
    def has_target(self) -> bool:
        return self.target is not None

    def __len__(self) -> int:
        return len(self.source)



"""Command-line interface."""

import logging

from lightning.pytorch import callbacks as pytorch_callbacks, cli
import omegaconf

from .. import callbacks, data, models, trainers


# Register OmegaConf resolvers here.
#
# This allows expressions of the form:
#
# hidden_size: ${multiply:${model.init_args.embedding_size}, 4}
#
# which means that the hidden size is 4x the embedding size.
omegaconf.OmegaConf.register_new_resolver("multiply", lambda x, y: x * y)


class YoyodyneCLI(cli.LightningCLI):
    """The Yoyodyne CLI interface.

    Use with `--help` to see the full list of options.
    """

    def add_arguments_to_parser(
        self, parser: cli.LightningArgumentParser
    ) -> None:
        parser.add_lightning_class_args(
            pytorch_callbacks.ModelCheckpoint,
            "checkpoint",
            required=False,
        )
        parser.add_lightning_class_args(
            callbacks.PredictionWriter,
            "prediction",
            required=False,
        )
        parser.link_arguments(
            "data.target_vocab_size",
            "model.init_args.target_vocab_size",
            apply_on="instantiate",
        )
        # Just needed for the transducer to create the expert.
        parser.link_arguments(
            "data.index",
            "model.init_args.index",
            apply_on="instantiate",
        )
        parser.link_arguments(
            "data.target_sep",
            "prediction.target_sep",
            apply_on="instantiate",
        )
        parser.link_arguments(
            "data.vocab_size",
            "model.init_args.vocab_size",
            apply_on="instantiate",
        )
        parser.link_arguments(
            "data.model_dir",
            "trainer.logger.init_args.save_dir",
            apply_on="instantiate",
        )


def main() -> None:
    logging.basicConfig(
        format="%(levelname)s: %(asctime)s - %(message)s",
        datefmt="%d-%b-%y %H:%M:%S",
        level="INFO",
    )
    # Select the model.
    YoyodyneCLI(
        models.BaseModel,
        data.DataModule,
        parser_kwargs={"parser_mode": "omegaconf"},
        save_config_callback=None,
        subclass_mode_model=True,
        # Prevents predictions from accumulating in memory; see the
        # documentation in `trainers.py` for more context.
        trainer_class=trainers.Trainer,
    )


def python_interface(args: cli.ArgsType = None) -> None:
    """Interface to use models through Python."""
    YoyodyneCLI(
        models.BaseModel,
        data.DataModule,
        parser_kwargs={"parser_mode": "omegaconf"},
        save_config_callback=None,
        subclass_mode_model=True,
        # See above for explanation.
        trainer_class=trainers.Trainer,
        args=args,
    )



"""Runs a W&B sweep."""

import argparse
import functools
import logging
import subprocess
import sys
import tempfile
import traceback
import warnings
from typing import Any, Dict, List, TextIO

import wandb
import yaml

from .. import util

warnings.filterwarnings("ignore", ".*is a wandb run already in progress.*")


def train_sweep(
    config: Dict[str, Any],
    temp_config: TextIO,
    argv: List[str],
) -> None:
    """Runs a single training run.

    Args:
        config: config dictionary.
        temp_config: temporary configuration file handle.
        argv: command-line arguments.
    """
    populate_config(config, temp_config)
    run_sweep(argv)


def run_sweep(argv: List[str]) -> None:
    """Actually runs the sweep.

    Args:
        argv: command-line arguments.

    We encapsulate each run by using a separate subprocess, which ought to
    ensure that memory is returned (etc.).
    """
    process = subprocess.Popen(argv, stderr=subprocess.PIPE, text=True)
    for line in process.stderr:
        logging.info(line.rstrip())
    wandb.finish(exit_code=process.wait())


def populate_config(
    config: Dict[str, Any],
    temp_config_handle: TextIO,
) -> None:
    """Populates temporary configuration file.

    The wandb config data used here comes from the environment.

    Args:
        config: config dictionary.
        temp_config_handle: temporary configuration file handle.
    """
    wandb.init()
    for key, value in wandb.config.items():
        util.recursive_insert(config, key, value)
    yaml.safe_dump(config, temp_config_handle)


def main() -> None:
    logging.basicConfig(
        format="%(levelname)s: %(asctime)s - %(message)s",
        datefmt="%d-%b-%y %H:%M:%S",
        level="INFO",
    )
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--sweep_id", required=True, help="ID for the sweep.")
    parser.add_argument(
        "--entity", required=True, help="The entity scope for the project."
    )
    parser.add_argument(
        "--project", required=True, help="The project of the sweep."
    )
    parser.add_argument(
        "--command",
        default="yoyodyne",
        help="Command to use when running sweep (default: %(default)s).",
    )
    parser.add_argument("--count", type=int, help="Number of runs to perform.")
    parser.add_argument("--config", required=True)
    # We pass the known args to main but remove them from ARGV.
    # See: https://docs.python.org/3/library/argparse.html#partial-parsing
    # This allows the user to override config arguments with CLI arguments.
    args, sys.argv[1:] = parser.parse_known_args()
    config = util.load_config(args.config)
    # TODO: Consider enabling the W&B logger; we are not sure if things will
    # unless this is configured.
    temp_config = tempfile.NamedTemporaryFile(mode="w", suffix=".yaml")
    argv = [
        args.command,
        "fit",
        "--config",
        temp_config.name,
        *sys.argv[1:],
    ]
    try:
        wandb.agent(
            sweep_id=args.sweep_id,
            entity=args.entity,
            project=args.project,
            function=functools.partial(train_sweep, config, temp_config, argv),
            count=args.count,
        )
    except Exception:
        # Exits gracefully, so W&B logs the error.
        logging.fatal(traceback.format_exc())
        wandb.finish(exit_code=1)
        exit(1)



"""Default values for flags and modules."""

import math

from . import optimizers, schedulers

# All elements should be styled as CONSTANTS.

# Scalar constants.
EPSILON = 1e-7
NEG_LOG_EPSILON = -math.log(EPSILON)
NEG_INF = -math.inf

# Default text encoding.
ENCODING = "utf-8"

# Data configuration arguments.
SOURCE_COL = 1
TARGET_COL = 2
FEATURES_COL = 0
SOURCE_SEP = ""
TARGET_SEP = ""
FEATURES_SEP = ";"
TIED_VOCABULARY = True

# Architecture arguments.
ATTENTION_CONTEXT = 0
ATTENTION_HEADS = 4
BIDIRECTIONAL = True
EMBEDDING_SIZE = 128
ENFORCE_MONOTONIC = False
HIDDEN_SIZE = 512
LAYERS = 1
MAX_LENGTH = 128
TIE_EMBEDDINGS = True

# Training arguments.
BATCH_SIZE = 32
BETA1 = 0.9
BETA2 = 0.999
DROPOUT = 0.2
LABEL_SMOOTHING = 0.0
ORACLE_FACTOR = 1
OPTIMIZER = optimizers.Adam
SCHEDULER = schedulers.Dummy

# Decoding arguments.
BEAM_WIDTH = 1
TEACHER_FORCING = True

# Extra evaluation metrics.
EVAL_METRICS = set()



"""Generation probability modules for pointer-generators."""

import math

import torch
from torch import nn


class GenerationProbability(nn.Module):
    """Generation probability for a pointer-generator.

    Args:
        embedding_size (int): embedding dimensions.
        hidden_size (int): decoder hidden state dimensions.
        attention_size (int): dimensions of combined encoder attentions.
    """

    stdev = 1 / math.sqrt(100)

    W_attention: nn.Linear
    W_hs: nn.Linear
    W_emb: nn.Linear
    bias: nn.Parameter

    def __init__(
        self,
        embedding_size: int,
        hidden_size: int,
        attention_size: int,
    ):
        super().__init__()
        self.W_emb = nn.Linear(embedding_size, 1, bias=False)
        self.W_hs = nn.Linear(hidden_size, 1, bias=False)
        self.W_attention = nn.Linear(attention_size, 1, bias=False)
        self.bias = nn.Parameter(torch.Tensor(1))
        self.bias.data.uniform_(-self.stdev, self.stdev)

    def forward(
        self,
        attention_context: torch.Tensor,
        decoder_hiddens: torch.Tensor,
        target_embeddings: torch.Tensor,
    ) -> torch.Tensor:
        """Computes the generation probability

        This is a function of the context vectors, decoder hidden states, and
        target embeddings, where each is first mapped to a scalar value by a
        learnable weight matrix.

        Args:
            attention_context (torch.Tensor): combined context vector over
                source and features of shape
                B x sequence_length x attention_size.
            decoder_hiddens (torch.Tensor): decoder hidden state of shape
                B x sequence_length x hidden_size.
            target_embeddings (torch.Tensor): decoder input of shape
                B x sequence_length x embedding_size.

        Returns:
            torch.Tensor: generation probability of shape B.
        """
        # -> B x sequence_length x 1.
        p_gen = self.W_attention(attention_context) + self.W_hs(
            decoder_hiddens
        )
        p_gen += self.W_emb(target_embeddings) + self.bias.expand(
            attention_context.size(0), 1, -1
        )
        # -> B x 1 x sequence_length.
        return torch.sigmoid(p_gen)



"""Yoyodyne models."""

from .. import defaults  # noqa: F401
from .base import BaseModel  # noqa: F401
from .base import ConfigurationError  # noqa: F401
from .hard_attention import HardAttentionGRUModel  # noqa: F401
from .hard_attention import HardAttentionLSTMModel  # noqa: F401
from .pointer_generator import PointerGeneratorGRUModel  # noqa: F401
from .pointer_generator import PointerGeneratorLSTMModel  # noqa: F401
from .pointer_generator import PointerGeneratorTransformerModel  # noqa: F401
from .rnn import GRUModel  # noqa: F401
from .rnn import LSTMModel  # noqa: F401
from .rnn import SoftAttentionGRUModel  # noqa: F401
from .rnn import SoftAttentionLSTMModel  # noqa: F401
from .transducer import TransducerGRUModel  # noqa: F401
from .transducer import TransducerLSTMModel  # noqa: F401
from .transformer import TransformerModel  # noqa: F401



"""Yoyodyne modules."""

from .attention import Attention  # noqa: F401
from .base import BaseModule  # noqa: F401
from .generation_probability import GenerationProbability  # noqa: F401
from .hard_attention import ContextHardAttentionGRUDecoder  # noqa: F401
from .hard_attention import ContextHardAttentionLSTMDecoder  # noqa: F401
from .hard_attention import HardAttentionRNNDecoder  # noqa: F401
from .hard_attention import HardAttentionGRUDecoder  # noqa: F401
from .hard_attention import HardAttentionLSTMDecoder  # noqa: F401
from .linear import LinearEncoder  # noqa: F401
from .rnn import GRUDecoder  # noqa: F401
from .rnn import GRUEncoder  # noqa: F401
from .rnn import LSTMDecoder  # noqa: F401
from .rnn import LSTMEncoder  # noqa: F401
from .rnn import RNNDecoder  # noqa: F401
from .rnn import RNNState  # noqa: F401
from .rnn import SoftAttentionGRUDecoder  # noqa: F401
from .rnn import SoftAttentionLSTMDecoder  # noqa: F401
from .transformer import FeatureInvariantTransformerEncoder  # noqa: F401
from .transformer import TransformerDecoder  # noqa: F401
from .transformer import TransformerEncoder  # noqa: F401
from .transformer import TransformerPointerDecoder  # noqa: F401



"""Data classes."""

from .batches import Batch, PaddedTensor  # noqa: F401
from .datamodules import DataModule  # noqa: F401
from .datasets import Dataset  # noqa: F401
from .indexes import Index  # noqa: F401
from .mappers import Mapper  # noqa: F401
from .tsv import TsvParser  # noqa: F401



"""Hard monotonic neural HMM classes."""

import abc
from typing import Optional, Tuple

import torch
from torch import nn

from .. import data, defaults, special
from . import base, embeddings, modules


class HardAttentionRNNModel(base.BaseModel):
    """Abstract base class for hard attention models.

    Learns probability distribution of target string by modeling transduction
    of source string to target string as Markov process. Assumes each symbol
    produced is conditioned by state transitions over each source symbol.

    The default model assumes independence between state and non-monotonic
    progression over source string. `enforce_monotonic` enforces monotonic
    state transition (model progresses over each source symbol), and
    attention context allows conditioning of state transition over the
    previous n states.

    If features are provided, the encodings are fused by concatenation of the
    source encoding with the features encoding, averaged across the length
    dimension and then scattered along the source length dimension, on the
    encoding dimension.

    After:
        Wu, S. and Cotterell, R. 2019. Exact hard monotonic attention for
        symbol-level transduction. In _Proceedings of the 57th Annual
        Meeting of the Association for Computational Linguistics_, pages
        1530-1537.

    Original implementation:
        https://github.com/shijie-wu/neural-transducer

     Args:
        *args: passed to superclass.
        attention_context (int, optional): size of context window for
            conditioning state transition; if 0, state transitions are
            independent.
        enforce_monotonic (bool, optional): enforces monotonic state
            transition in decoding.
        **kwargs: passed to superclass.
    """

    enforce_monotonic: bool
    attention_context: int

    def __init__(
        self,
        *args,
        attention_context: int = defaults.ATTENTION_CONTEXT,
        enforce_monotonic: bool = defaults.ENFORCE_MONOTONIC,
        **kwargs,
    ):
        self.attention_context = attention_context
        self.enforce_monotonic = enforce_monotonic
        super().__init__(*args, **kwargs)
        self.classifier = nn.Linear(
            self.decoder.output_size, self.target_vocab_size
        )

    # Prevents a loss function object from being constructed.

    def _get_loss_func(self) -> None:
        return

    def _loss(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        """Computes average loss.

        Args:
            encoded (torch.Tensor): encoded source symbols
                of shape B x src_len x (encoder_hidden * num_directions).
            mask (torch.Tensor): mask.
            target (torch.Tensor): target symbols.

        Returns:
            torch.Tensor: average loss.
        """
        batch_size = mask.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        emissions, transitions, state = self.decode_step(
            encoded, mask, symbol, state
        )
        likelihood = self._emissions(
            transitions[:, 0].unsqueeze(1),
            emissions,
            target[:, 0],
        )
        target_len = target.size(1)
        for t in range(1, target_len):
            emissions, transitions, state = self.decode_step(
                encoded, mask, target[:, t - 1].unsqueeze(1), state
            )
            likelihood = self._transitions(likelihood, transitions)
            likelihood = self._emissions(likelihood, emissions, target[:, t])
        return self._scale(likelihood, target_len)

    def decode_step(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: modules.RNNState,
    ) -> Tuple[torch.Tensor, torch.Tensor, modules.RNNState]:
        """Single decoder step.

        Args:
            encoded (torch.Tensor): encoded source symbols of shape
                B x src_len x (encoder_hidden * num_directions).
            mask (torch.Tensor): mask.
            symbol (torch.Tensor): target symbol for current state.
            state (modules.RNNState): RNN state.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: emission
                probabilities, transition probabilities, and the RNN state.
        """
        emissions, transitions, state = self.decoder(
            encoded,
            mask,
            symbol,
            state,
            self.embeddings,
        )
        logits = self.classifier(emissions)
        scores = nn.functional.log_softmax(logits, dim=2)
        # Expands matrix for all time steps.
        if self.enforce_monotonic:
            transitions = self._apply_mono_mask(transitions)
        return scores, transitions, state

    @staticmethod
    def _apply_mono_mask(transitions: torch.Tensor) -> torch.Tensor:
        """Applies monotonic attention mask to transition probabilities.

        Enforces zero log-probability values for all non-monotonic relations
        in the transition tensor (i.e., all values i < j per row j).

        Args:
            transitions (torch.Tensor): transition probabilities between
                all hidden states of shape B x src_len x src_len.

        Returns:
            torch.Tensor: masked transition probabilities of shape
                B x src_len x src_len.
        """
        mask = torch.ones_like(transitions[0]).triu().unsqueeze(0)
        # Using 0 log-probability value for masking; this is borrowed from the
        # original implementation.
        mask = (mask - 1) * defaults.NEG_LOG_EPSILON
        transitions = transitions + mask
        transitions = transitions - transitions.logsumexp(dim=2, keepdim=True)
        return transitions

    def init_embeddings(
        self,
        num_embeddings: int,
        embedding_size: int,
    ) -> nn.Embedding:
        """Initializes the embedding layer.

        Args:
            num_embeddings (int): number of embeddings.
            embedding_size (int): dimension of embeddings.

        Returns:
            nn.Embedding: embedding layer.
        """
        return embeddings.normal_embedding(num_embeddings, embedding_size)

    def forward(self, batch: data.Batch) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]: if
                training, the loss tensor; otherwise, also predictions of shape
                B x pred_seq_length.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Feature column specified but no feature
                encoder specified.
        """
        encoded = self.source_encoder(batch.source, self.embeddings)
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            features_encoded = features_encoded.mean(dim=1, keepdim=True)
            features_encoded = features_encoded.expand(-1, encoded.size(1), -1)
            encoded = torch.cat((encoded, features_encoded), dim=2)
        elif batch.has_features:
            raise base.ConfigurationError(
                "Feature column specified but no feature encoder specified"
            )
        if self.training:
            return self._loss(
                encoded,
                batch.source.mask,
                batch.target.padded,
            )
        else:
            return self.greedy_decode(
                encoded,
                batch.source.mask,
                batch.target.padded if batch.has_target else None,
            )

    def greedy_decode(
        self,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        target: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Decodes a sequence given the encoded input.

        Decodes until all sequences in a batch have reached END up to a
        specified length depending on the `target` args.

        Args:
            source_encoded (torch.Tensor): encoded source symbols of shape
                B x src_len x (encoder_hidden * num_directions)
            source_mask (torch.Tensor): mask.
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues until this length is reached.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: the loss and predictions of
                shape B x pred_seq_len.
        """
        batch_size = source_mask.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        emissions, transitions, state = self.decode_step(
            source_encoded, source_mask, symbol, state
        )
        likelihood = transitions[:, 0].unsqueeze(1)
        symbol = self._greedy_step(likelihood, emissions)
        predictions = [symbol]
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        # We already did one step.
        for t in range(max_num_steps - 1):
            emissions, transitions, state = self.decode_step(
                source_encoded,
                source_mask,
                symbol.unsqueeze(1),
                state,
            )
            likelihood = self._transitions(likelihood, transitions)
            symbol = self._greedy_step(likelihood, emissions)
            predictions.append(symbol)
            if target is None:
                final = torch.logical_or(final, symbol == special.END_IDX)
                if final.all():
                    break
            likelihood = self._emissions(likelihood, emissions, symbol)
        # -> B x seq_len.
        predictions = torch.stack(predictions, dim=1)
        return self._scale(likelihood, t + 1), predictions

    @classmethod
    def _emissions(
        cls,
        likelihood: torch.Tensor,
        emissions: torch.Tensor,
        symbol: torch.Tensor,
    ) -> torch.Tensor:
        """Adds emission to likelihood.

        Args:
            likelihood (torch.Tensor): probabilities of shape
                B x 1 x src_len.
            emissions (torch.Tensor): slice of emission probabilities of shape
                B x src_len x vocab_size.
            symbol (torch.Tensor): slice of symbols of shape B.

        Returns:
            torch.Tensor: probabilities of shape B x 1 x src_len.
        """
        return likelihood + cls._gather_at_idx(emissions, symbol)

    @staticmethod
    def _gather_at_idx(
        emissions: torch.Tensor, symbol: torch.Tensor
    ) -> torch.Tensor:
        """Collects probability of the symbol across all states.

        To calculate the final emission probability, the pseudo-HMM
        graph needs to aggregate the final emission probabilities of
        target symbols across all potential hidden states in the emissions.

        Args:
            emissions (torch.Tensor): slice of emission probabilities of
                shape B x src_len.
            symbol (torch.Tensor): target symbol to poll probabilities for
                shape B.

        Returns:
            torch.Tensor: emission probabilities of the symbol for each hidden
                state, of size B x 1 x src_len.
        """
        batch_size = emissions.size(0)
        src_seq_len = emissions.size(1)
        idx = symbol.view(-1, 1).expand(batch_size, src_seq_len).unsqueeze(-1)
        output = torch.gather(emissions, -1, idx).view(
            batch_size, 1, src_seq_len
        )
        idx = idx.view(batch_size, 1, src_seq_len)
        pad_mask = (idx != special.PAD_IDX).float()
        return output * pad_mask

    @staticmethod
    def _greedy_step(
        likelihood: torch.Tensor,
        emissions: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Greedy decoding of current timestep.

        Args:
            likelihood (torch.Tensor): probabilities of shape B x 1 x src_len.
            emissions (torch.Tensor): slice of emission probabilities of shape
                B x src_len x vocab_size.

        Returns:
            torch.Tensor: symbols of shape B.
        """
        likelihood = likelihood + emissions.transpose(1, 2)
        return torch.argmax(likelihood.logsumexp(dim=2), dim=1)

    @staticmethod
    def _scale(likelihood: torch.Tensor, size: int) -> torch.Tensor:
        """Scales likelihood to compute loss.

        Args:
            likelihood (torch.Tensor): probailities of shape
                B x 1 x src_len.
            size (int).

        Returns:
            torch.Tensor: scaled loss.
        """
        return -likelihood.logsumexp(dim=2).mean() / size

    @staticmethod
    def _transitions(
        likelihood: torch.Tensor, transitions: torch.Tensor
    ) -> torch.Tensor:
        """Adds transitions to likelihood.

        Args:
            likelihood (torch.Tensor): probabilities of shape
                B x 1 x src_len.
            transitions (torch.Tensor): slice of transition probabilities
                of shape B x src_len x src_len.

        Returns:
            torch.Tensor: probabilities of shape B x 1 x src_len.
        """
        return torch.logsumexp(
            likelihood + transitions.transpose(1, 2),
            dim=2,
            keepdim=True,
        ).transpose(1, 2)

    def predict_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        _, predictions = self(batch)
        return predictions

    def test_step(self, batch: data.Batch, batch_idx: int) -> None:
        _, predictions = self(batch)
        self._update_metrics(predictions, batch.target.padded)

    def training_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        loss = self(batch)
        self.log(
            "train_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: data.Batch, batch_idx: int) -> None:
        loss, predictions = self(batch)
        self.log(
            "val_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            prog_bar=True,
        )
        self._update_metrics(predictions, batch.target.padded)

    @property
    def decoder_input_size(self) -> int:
        # We concatenate along the encoding dimension.
        if self.has_features_encoder:
            return (
                self.source_encoder.output_size
                + self.features_encoder.output_size
            )
        else:
            return self.source_encoder.output_size

    @abc.abstractmethod
    def get_decoder(self) -> modules.HardAttentionRNNDecoder: ...

    @property
    @abc.abstractmethod
    def name(self) -> str: ...


class HardAttentionGRUModel(HardAttentionRNNModel):
    """Hard attention with GRU backend."""

    def get_decoder(self):
        if self.attention_context > 0:
            return modules.ContextHardAttentionGRUDecoder(
                attention_context=self.attention_context,
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )
        else:
            return modules.HardAttentionGRUDecoder(
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )

    @property
    def name(self) -> str:
        return "hard attention GRU"


class HardAttentionLSTMModel(HardAttentionRNNModel):
    """Hard attention with LSTM backend."""

    def get_decoder(self):
        if self.attention_context > 0:
            return modules.ContextHardAttentionLSTMDecoder(
                attention_context=self.attention_context,
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                hidden_size=self.decoder_hidden_size,
                embedding_size=self.embedding_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )
        else:
            return modules.HardAttentionLSTMDecoder(
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )

    @property
    def name(self) -> str:
        return "hard attention LSTM"



"""Symbol index."""

from __future__ import annotations

import itertools
import pickle
from typing import Dict, Iterable, List, Optional

from torch import serialization
import yaml

from .. import defaults, special, util


class Error(Exception):
    pass


class Index:
    """Maintains the index over the vocabularies.

    Args:
        source_vocabulary (Iterable[str]).
        features_vocabulary (Iterable[str], optional).
        target_vocabulary (Iterable[str], optional).
        tie_embeddings (bool).
    """

    source_vocabulary: List[str]
    features_vocabulary = Optional[List[str]]
    target_vocabulary: List[str]
    tie_embeddings: bool
    _index2symbol: List[str]
    _symbol2index: Dict[str, int]

    def __init__(
        self,
        *,
        source_vocabulary: Iterable[str],
        features_vocabulary: Optional[Iterable[str]] = None,
        target_vocabulary: Optional[Iterable[str]] = None,
        tie_embeddings: bool = defaults.TIE_EMBEDDINGS,
    ):
        super().__init__()
        self.tie_embeddings = tie_embeddings
        # We store vocabularies separately for logging purposes.
        self.source_vocabulary = sorted(source_vocabulary)
        self.features_vocabulary = (
            sorted(features_vocabulary) if features_vocabulary else None
        )
        self.target_vocabulary = (
            sorted(target_vocabulary) if target_vocabulary else None
        )
        if self.tie_embeddings:
            # Vocabulary is the union of source and target.
            vocabulary = sorted(
                frozenset(
                    itertools.chain(source_vocabulary, target_vocabulary)
                )
            )
        else:
            # Vocabulary consists of target symbols followed by source symbols.
            vocabulary = self.target_vocabulary + self.source_vocabulary
        # FeatureInvariantTransformer assumes that features_vocabulary is at
        # the end of the vocabulary.
        if self.features_vocabulary is not None:
            vocabulary.extend(self.features_vocabulary)
        # Keeps special.SPECIAL first to maintain overlap with features.
        self._index2symbol = special.SPECIAL + vocabulary
        self._symbol2index = {c: i for i, c in enumerate(self._index2symbol)}

    def __len__(self) -> int:
        return len(self._index2symbol)

    # Lookup.

    def __call__(self, lookup: str) -> int:
        """Looks up index by symbol.

        Args:
            symbol (str).

        Returns:
            int.
        """
        return self._symbol2index.get(lookup, special.UNK_IDX)

    def get_symbol(self, index: int) -> str:
        """Looks up symbol by index.

        Args:
            index (int).

        Returns:
            str.
        """
        return self._index2symbol[index]

    # Serialization.

    @classmethod
    def read(cls, model_dir: str) -> Index:
        """Loads index.

        Args:
            model_dir (str).

        Returns:
            Index.
        """
        with open(cls.path(model_dir), "rb") as source:
            return pickle.load(source)

    def write(self, model_dir: str) -> None:
        """Writes index.

        Args:
            model_dir (str).
        """
        path = self.path(model_dir)
        util.mkpath(path)
        with open(path, "wb") as sink:
            pickle.dump(self, sink)

    @staticmethod
    def path(model_dir: str) -> str:
        """Computes path for the index file.

        Args:
            model_dir (str).

        Returns:
            str.
        """
        return f"{model_dir}/index.pkl"

    @staticmethod
    def _yaml_representer(dumper: yaml.Representer, data: Index):
        return dumper.represent_mapping(
            "!Index",
            {
                "source_vocabulary": data.source_vocabulary,
                "features_vocabulary": data.features_vocabulary,
                "target_vocabulary": data.target_vocabulary,
                "tie_embeddings": data.tie_embeddings,
            },
        )

    @staticmethod
    def _yaml_constructor(loader: yaml.Constructor, node: yaml.Node):
        node_value = loader.construct_mapping(node, deep=True)
        return Index(
            source_vocabulary=node_value.get("source_vocabulary"),
            features_vocabulary=node_value.get("features_vocabulary"),
            target_vocabulary=node_value.get("target_vocabulary"),
            tie_embeddings=node_value.get("tie_embeddings"),
        )

    # Properties.

    @property
    def symbols(self) -> List[str]:
        return list(self._symbol2index.keys())

    @property
    def vocab_size(self) -> int:
        return len(self._symbol2index)

    @property
    def source_vocab_size(self) -> int:
        if self.tie_embeddings:
            return self.vocab_size
        else:
            return len(self.SPECIAL) + len(self.source_vocabulary)

    @property
    def features_vocab_size(self) -> int:
        return len(self.features_vocabulary) if self.features_vocabulary else 0

    @property
    def target_vocab_size(self) -> int:
        if self.tie_embeddings:
            return self.vocab_size
        elif self.target_vocabulary:
            return len(special.SPECIAL) + len(self.target_vocabulary)
        else:
            return 0


# This whitelists the Index for safe serialization.


serialization.add_safe_globals([Index])
yaml.add_representer(Index, Index._yaml_representer, Dumper=yaml.SafeDumper)
yaml.add_constructor("!Index", Index._yaml_constructor, Loader=yaml.SafeLoader)



"""Custom schedulers."""

import math
from typing import List

from torch import optim


class Dummy(optim.lr_scheduler.LRScheduler):
    """A dummy scheduler that holds learning rate constant.

    Args:
        optimizer: optimizer.
    """

    def __init__(self, optimizer: optim.Optimizer):
        super().__init__(optimizer)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.optimizer})"

    def get_lr(self) -> List[float]:
        return [group["lr"] for group in self.optimizer.param_groups]


class WarmupInverseSquareRoot(optim.lr_scheduler.LambdaLR):
    """Linear warmup and then inverse square root decay.

    Linearly increases learning rate from 0 to the learning rate over the
    warmup epochs, then decreases learning rate according to an inverse root
    square schedule.

    After:
        Wu, S., Cotterell, R., and Hulden, M. 2021. Applying the transformer to
        character-level transductions. In Proceedings of the 16th Conference of
        the European Chapter of the Association for Computational Linguistics:
        Main Volume, pages 1901-1907.

    Args:
        optimizer (optim.Optimizer): optimizer.
        warmup_epochs (int): number of warmup epochs.
        *args: ignored.
        **kwargs: ignored.
    """

    warmup_epochs: int
    decay_factor: float

    def __init__(
        self,
        optimizer: optim.Optimizer,
        warmup_epochs: int,
        *args,
        **kwargs,
    ):
        self.warmup_epochs = warmup_epochs
        self.decay_factor = math.sqrt(warmup_epochs)
        super().__init__(optimizer, self.lr_lambda)

    def lr_lambda(self, epoch: int) -> float:
        """Computes the learning rate lambda at a given epoch.

        Args:
            epoch (int): current epoch.

        Returns:
            float: lr_lambda.
        """
        if epoch < self.warmup_epochs:
            # +1 in numerator avoids a zero-LR first epoch.
            return (epoch + 1) / self.warmup_epochs
        # +1 in base of exponent avoids an undefined operation (0 to a negative
        # exponent) in the unlikely case one is using this without warmup.
        return self.decay_factor * (epoch + 1) ** -0.5



"""Subclasses the trainer to prevent a memory leak in prediction.

This is adapted from UDTube:

    https://github.com/CUNY-CL/udtube/blob/master/udtube/trainers.py
"""

# TODO: replace the version in Yoyodyne-Pretrained.


from lightning.pytorch import trainer


class Trainer(trainer.Trainer):

    def predict(self, *args, **kwargs):
        return super().predict(*args, return_predictions=False, **kwargs)



"""Validation metrics.

The computation of loss is built into the models. A slight modification of
a built-in class from torchmetrics is used to compute exact match accuracy. A
novel symbol error rate (SER) implementation is also provided.

Adding additional metrics is relatively easy, though there are a lot of steps.
Suppose one wants to add a metric called Wham. Then one must:

* Implement `Wham(torchmetrics.Metric)` in this module.
* Add the following to the `BaseModel` in `models/base.py`:
    - add `wham: Optional[metrics.Wham]` to the member type declarations
    - add `compute_wham=False` to the constructor's arguments
    - add `self.wham = metric.Wham(...) if compute_wham else None` to the
      body of the constructor
    - add the following property:

        @property
        def has_wham(self) -> bool:
            return self.wham is not None

    - add the following to the body of `_reset_metrics`:

        if self.has_wham:
            self.wham.reset()

    - add the following to the body of `_update_metrics`:

        if self.has_wham:
            self.wham.update(predictions, target)

    - add the following to the body of `_log_metrics_on_epoch_end`:

        if self.has_wham:
            self.log(
                f"{subset}_wham",
                self.wham.compute(),
                logger=True,
                on_epoch=True,
                prog_bar=True,
            )
"""

import numpy
import torch
import torchmetrics

from . import special


class Error(Exception):
    pass


class Accuracy(torchmetrics.classification.MulticlassExactMatch):
    """Exact match string accuracy ignoring padding symbols."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, ignore_index=special.PAD_IDX, **kwargs)


class SER(torchmetrics.Metric):
    r"""Defines a symbol error rate metric.

    Symbol error rate is essentially minimum edit distance, in "symbols" (not
    characters, as in the `torchmetrics.text` module) scaled by the length
    of the gold hypotheses. Theoretically its range is $[0, \infty]$ but in
    practice it is usually in [0, 1]; smaller is better.

    Some definitions multiple this by 100; we don't bother here.

    We assume tensors of shape B x seq_len as input. For reasons documented
    below, seq_len must be $< 2^16$.

    This is intended to be a corpus-level statistic, so the number of edits and
    the lengths of strings are stored separately and are only combined as
    needed.

    It is not obvious whether the dynamic programming table ought to be stored
    on CPU, using a Numpy array, or in a 2d tensor on the accelerator. Since
    there is no need to track gradients and since it is accessed in a way that
    is likely to make good use of a CPU cache, the current implementation
    assumes the former, but this can be re-evaluated in light of profiling.

    One can imagine imposing different user-specified costs for the different
    edit operations, but rule this out for YAGNI reasons.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_state("edits", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("length", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, hypo: torch.Tensor, gold: torch.Tensor) -> None:
        """Accumulates edit distance sufficient statistics for a batch.

        This also performs all the necessary data validation work:

        Args:
            hypo (torch.Tensor): a tensor of hypothesis data of shape
                B x target_vocab_size x seq_len or B x seq_len.
            gold (torch.Tensor): a tensor of gold data of shape B x seq_len.

        Raises:
            Error: Hypothesis tensor is not 2d or 3d.
            Error: Gold tensor is not 2d.
            Error: Hypothesis and gold batch sizes do not match.
            Error: Hypothesis string lengths exceeds precision.
            Error: Gold string lengths exceeds precision.
        """
        if hypo.ndim < 2 or hypo.ndim > 3:
            raise Error(f"Hypothesis tensor is not 2d or 3d ({hypo.ndim})")
        if hypo.ndim == 3:
            hypo = torch.argmax(hypo, dim=1)
        if gold.ndim != 2:
            raise Error(f"Gold tensor is not 2d ({gold.ndim})")
        if hypo.size(0) != gold.size(0):
            raise Error(
                "Hypothesis and gold batch sizes do not match "
                f"({gold.size(0)} != {hypo.size(0)})"
            )
        # uint16 is used for the dynamic programming table, so this
        # implementation is not necessarily correct for strings longer than
        # $2^16 = 65536$. This is not much of a limitation in practice because
        # quadratic growth makes the computation infeasible at that length
        # anyways. This checks the length of the second dimension to ensure it
        # does not exceed this length.
        max_size = numpy.iinfo(numpy.uint16).max
        if hypo.size(-1) > max_size:
            raise Error(
                "Hypothesis string lengths exceeds precision "
                f"({hypo.size(-1)} > {max_size})"
            )
        if gold.size(-1) > max_size:
            raise Error(
                "Gold string lengths exceeds precision "
                f"({gold.size(-1)} > {max_size})"
            )
        # Iterates over every element in batch.
        for hypo_row, gold_row in zip(hypo, gold):
            self._row_edit_distance(hypo_row, gold_row)

    def _row_edit_distance(
        self,
        hypo: torch.Tensor,
        gold: torch.Tensor,
    ) -> None:
        """Computes edit distance sufficient statistics for single tensors.

        This makes the following assumptions about the input tensors:

        * They are 1d and can be interpreted as single strings.
        * The end of the string is delimited by END_IDX; all indices before
          the first END_IDX are part of the string and all indices after it
          are not.

        Args:
            hypo (torch.Tensor): hypothesis 1d tensor.
            gold (torch.Tensor): gold 1d tensor.
        """
        # The - 1 term reflects that `END_IDX` is not part of the string
        # with respect to edit distance. This also cannot fail.
        gold_length = torch.nonzero(gold == special.END_IDX)[0].item() - 1
        self.length += gold_length
        try:
            hypo_length = torch.nonzero(hypo == special.END_IDX)[0].item() - 1
        except IndexError:
            hypo_length = -1
        if hypo_length < 0:
            # If END_IDX isn't present, we'll consider this is a "total loss"
            # with an edit distance equivalent to the gold length. One can
            # imagine more elaborate strategies but this oughta do.
            self.edits += gold_length
            return
        table = numpy.zeros(
            (hypo_length + 1, gold_length + 1), dtype=numpy.uint16
        )
        table[:, 0] = range(hypo_length + 1)
        table[0, :] = range(gold_length + 1)
        for i in range(1, hypo_length + 1):
            for j in range(1, gold_length + 1):
                if hypo[i - 1] == gold[j - 1]:
                    table[i, j] = table[i - 1, j - 1]
                else:
                    table[i, j] = (
                        min(
                            table[i - 1, j],
                            table[i, j - 1],
                            table[i - 1, j - 1],
                        )
                        + 1
                    )
        self.edits += table[-1, -1]

    def compute(self) -> torch.Tensor:
        return self.edits / self.length



"""Collators and related utilities."""

import dataclasses
import logging

from typing import List

from .. import defaults
from . import batches, datasets


class Error(Exception):
    pass


@dataclasses.dataclass
class Collator:
    """Pads data."""

    has_features: bool
    has_target: bool
    max_source_length: int = defaults.MAX_LENGTH
    max_features_length: int = defaults.MAX_LENGTH
    max_target_length: int = defaults.MAX_LENGTH

    def _source_length_error(self, padded_length: int) -> None:
        """Callback function for excessive source length.

        Args:
            padded_length (int): The length of the the padded tensor.

        Raises:
            Error.
        """
        if padded_length > self.max_source_length:
            raise Error(
                f"The length of a source sample ({padded_length}) is greater "
                f"than the `--max_source_length` specified "
                f"({self.max_source_length})"
            )

    def _features_length_error(self, padded_length: int) -> None:
        """Callback function for excessive features length.

        Args:
            padded_length (int): The length of the the padded tensor.

        Raises:
            Error.
        """
        if padded_length > self.max_features_length:
            raise Error(
                f"The length of a features sample ({padded_length}) is "
                f"greater than the `--max_features_length` specified "
                f"({self.max_features_length})"
            )

    def _target_length_warning(self, padded_length: int) -> None:
        """Callback function for excessive target length.

        Since `max_target_length` just truncates during inference, this is
        simply a suggestion.

        Args:
            padded_length (int): The length of the the padded tensor.
        """
        if padded_length > self.max_target_length:
            logging.info(
                "The length of a batch (%d}) is greater than the "
                "max_target_length specified (%d). Sequences may be "
                "truncated during decoding. Consider increasing "
                "max_target_length.",
                padded_length,
                self.max_target_length,
            )

    def pad_source(
        self, itemlist: List[datasets.Item]
    ) -> batches.PaddedTensor:
        """Pads source.

        Args:
            itemlist (List[datasets.Item]).

        Returns:
            batches.PaddedTensor.
        """
        return batches.PaddedTensor(
            [item.source for item in itemlist],
            self._source_length_error,
        )

    def pad_features(
        self,
        itemlist: List[datasets.Item],
    ) -> batches.PaddedTensor:
        """Pads features.

        Args:
            itemlist (List[datasets.Item]).

        Returns:
            batches.PaddedTensor.
        """
        return batches.PaddedTensor([item.features for item in itemlist])

    def pad_target(
        self, itemlist: List[datasets.Item]
    ) -> batches.PaddedTensor:
        """Pads target.

        Args:
            itemlist (List[datasets.Item]).

        Returns:
            batches.PaddedTensor.
        """
        return batches.PaddedTensor(
            [item.target for item in itemlist],
            self._target_length_warning,
        )

    def __call__(self, itemlist: List[datasets.Item]) -> batches.Batch:
        """Pads all elements of an itemlist.

        Args:
            itemlist (List[datasets.Item]).

        Returns:
            batches.Batch.
        """
        return batches.Batch(
            self.pad_source(itemlist),
            self.pad_features(itemlist) if self.has_features else None,
            self.pad_target(itemlist) if self.has_target else None,
        )



"""Positional encoding modules for transformers."""

import math
from typing import Optional

import torch
from torch import nn

from ... import defaults, special


class PositionalEncoding(nn.Module):
    """Positional encoding.

    After:
        https://pytorch.org/tutorials/beginner/transformer_tutorial.html.

    Args:
        d_model (int).
        max_source_length (int, optional).
    """

    def __init__(
        self, d_model: int, max_source_length: int = defaults.MAX_LENGTH
    ):
        super().__init__()
        positional_encoding = torch.zeros(max_source_length, d_model)
        position = torch.arange(
            0, max_source_length, dtype=torch.float
        ).unsqueeze(1)
        scale_factor = -math.log(10000.0) / d_model
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * scale_factor
        )
        positional_encoding[:, 0::2] = torch.sin(position * div_term)
        positional_encoding[:, 1::2] = torch.cos(position * div_term)
        positional_encoding = positional_encoding.unsqueeze(0)
        self.register_buffer("positional_encoding", positional_encoding)

    def forward(
        self,
        symbol: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Computes the positional encoding.

        Args:
            symbol (torch.Tensor): symbol indices to encode.
            mask (torch.Tensor, optional): optional mask for positions not to
                be encoded.

        Returns:
            torch.Tensor: positional embedding.
        """
        out = self.positional_encoding.repeat(symbol.size(0), 1, 1)
        if mask is None:
            indices = torch.arange(symbol.size(1))
        else:
            # Indices should all be 0's until the first unmasked position.
            indices = torch.cumsum(mask, dim=1)
        # Selects the tensors from `out` at the specified indices.
        out = out[torch.arange(out.size(0)).unsqueeze(-1), indices]
        # Zeros out pads.
        out = out * symbol.ne(special.PAD_IDX).unsqueeze(2)
        return out
